import cocos
import pyglet
from pyglet import clock
from cocos.director import director
import random
import blood, vomit

class GameScene(cocos.scene.Scene):
    def __init__(self):
        super(GameScene, self).__init__()
        
        #security monitor
        self.foreground_layer = ForegroundLayer()
        #interface layer
        self.ui_layer = UILayer()
        #the game layer - where it all happens
        self.game_layer = GameLayer()
        
        self.add(self.game_layer)
        
class ForegroundLayer(cocos.layer.Layer):
    """
    Security console graphic, blood spatter
    """
    
    def __init__(self):
        super(ForegroundLayer, self).__init__()

class UILayer(cocos.layer.Layer):
    """
    Any interface elements..
    """
    
    def __init__(self):
        super(UILayer, self).__init__()
        
    def consoleMessage(self, message):
        """
        Output a message to the console
        """
        pass
        
class GameLayer(cocos.layer.Layer):
    """
    Where all the action is.
    """
    
    is_event_handler = True
    
    def __init__(self):
        super(GameLayer, self).__init__()
        
        self.keys = {'saw':pyglet.window.key.SPACE ,'direction':{pyglet.window.key.LEFT:'left', pyglet.window.key.RIGHT:'right', pyglet.window.key.UP:'up', pyglet.window.key.DOWN:'down'}}
        self.on_keys = {'direction':[]}
        self.game_on = True
        self.mutants = []
        
        #load level
        batch = cocos.batch.BatchNode()
        self.add(batch)
        
        self.level = Level()
        self.level.build(batch)
        
        #add player
        self.player = Player(self.level.start)
        self.add(self.player.sprite, 1)
        
        #add mutants
        for m in self.level.mutants:
            mutant = Mutant(m[0], m[1])
            self.add(mutant.sprite, 0)
            self.mutants.append(mutant)
        
        self.schedule(self.update)
        
    def update(self, dt):
        """
        Game loop
        @param dt: Delta, time since last update
        """
        
        if self.game_on:
            
            #act on presses in direction key stack
            if len(self.on_keys['direction']) > 0:
                
                #update movement animation sprite
                if self.on_keys['direction'][-1] != self.player.direction:
                    self.player.update_sprite(self.on_keys['direction'][-1])
                
                self.player.is_moving = True
            else:
                self.player.is_moving = False
            
            self.update_object(dt, self.player)
            
            #update mutants
            for mutant in self.mutants:
                
                #mutant getting sawed
                if self.player.saw:
                    
                    #TODO: Collission is just with the player right now...
                    if self.player.saw.state == 'on' and self.player.sprite.get_rect().intersects(mutant.sprite.get_rect()):
                        #get some blood down, yar!
                        b = blood.Blood(robot_direction = self.player.direction)
                        mutant.sprite.add(b, 2)
                        #TODO: take position of saw, mutnat direction into account, add some randomness
                        b.position = (0,0)
                
                mutant.ai(self.player.sprite.get_rect())
                self.update_object(dt, mutant)
            
            self.update_viewpoint()

    def update_object(self, dt, game_object):
        """
        Update the position of the Object based on its direction and speed
        """
        #TODO: acceleration. Otherwise we don't really need a speed property
        
        if game_object.is_moving:
            speed = game_object.directions[game_object.direction] * game_object.max_speed * dt
            position = game_object.position + speed
            x, y = (position + game_object.offset - game_object.base / 2)
            width, height = game_object.base
            
            collision_rectangle = cocos.rect.Rect(x, y, width, height)
            
            #check if we can move in that direction
            if self.level.is_open(collision_rectangle):
                game_object.speed = speed
                game_object.position = position
                game_object.sprite.position = position
                game_object.is_blocked = False

            else:
                #TODO: get closest open tile direction
                avoid_direction = self.level.direction_to_open_tile(game_object.direction, game_object.position + game_object.offset)
                
                if avoid_direction:
                    game_object.speed = game_object.directions[avoid_direction] * game_object.max_speed * dt
                    game_object.position = game_object.position + game_object.speed
                    game_object.sprite.position = game_object.position
                    game_object.is_blocked = False
                else:
                    game_object.speed = cocos.euclid.Point2(0, 0)
                    game_object.is_blocked = True
            
            #TODO: Collisions with other objects. E.g. player pushing enemies, enemies attacking player, bullets hitting player/enemies etc.
            
        else:
            #TODO:somehow stop movement animation
            pass
        
    def update_viewpoint(self):
        """
        Move the level around the player, or move the view and what not.
        """
        #TODO: implement view move to location   
        #if self.move_to_point:
        #    move..
        #else:
        self.position = self.translate_vewpoint_location(self.player.sprite.position)
    
    def translate_vewpoint_location(self, point):
        centerPoint = cocos.euclid.Point2(director.window.width / 2, director.window.height / 2)
        viewPoint = centerPoint - point
        
        return viewPoint
    
    def on_key_press(self, key, modifiers):
        keys = self.keys['direction']
        if key in keys:
            self.on_keys['direction'].append(keys[key])
            return True
        
        elif key == self.keys['saw']:
            self.player.saw_on()
            return True
        
        return False

    def on_key_release(self, key, modifiers):
        keys = self.keys['direction']
        if key in keys:
            self.on_keys['direction'].pop(0)
            return True
        
        elif key == self.keys['saw']:
            self.player.saw_off()
            return True
        return False


class GameObject(object):
    """
    Base for all Objects, player or enemy or bullet
    """
    
    def __init__(self, name, position = (0,0), active = True, direction = 'down', is_moving = False, animations = None):
        self.active = active
        self.direction = direction
        
        self.index = 0
        self.max_speed = 40
        self.is_moving = is_moving
        self.acceleration = 10
        self.is_blocked = False
        self.is_following = False
        self.is_attacking = False
        
        #in case some mutants/objects can only move left/right etc
        self.available_directions = ['left', 'right', 'up', 'down']
        
        #horizontal is a bit faster..
        self.directions = {'down': cocos.euclid.Point2(0, -1), 'up': cocos.euclid.Point2(0, 1), 'left': cocos.euclid.Point2(-1.2, 0), 'right': cocos.euclid.Point2(1.2, 0) }
        
        self.speed = cocos.euclid.Point2(0, 0)
        self.position = cocos.euclid.Point2(position[0], position[1])
        
        self.base = cocos.euclid.Point2(81, 28)
        self.offset = cocos.euclid.Point2(0, -26)
        
        if animations is None:
            #setup animations for each direction
            sheet = pyglet.image.load('%s.png' % name)
            sequence = pyglet.image.ImageGrid(sheet, 2, 4)
            
            self.animations = {}
            self.animations['down'] = pyglet.image.Animation.from_image_sequence([sequence[0], sequence[4]], 0.15)
            self.animations['up'] = pyglet.image.Animation.from_image_sequence([sequence[1], sequence[5]], 0.15)
            self.animations['left'] = pyglet.image.Animation.from_image_sequence([sequence[3], sequence[7]], 0.15)
            self.animations['right'] = pyglet.image.Animation.from_image_sequence([sequence[2], sequence[6]], 0.15)
        else:
            self.animations = animations
        
        
        self.sprite = cocos.sprite.Sprite(self.animations[self.direction])
        self.sprite.position = position
    
    def update_sprite(self, direction):
        self.direction = direction
        self.sprite.image = self.animations[direction]
        self.sprite.position = self.position
        
    def direction_to_point(self, point):
        """
        get the direction to go in to reach given point
        
        Go for the longer direction first
        """
        
        diff = [self.position[0] - point[0], self.position[1] - point[1]]
        
        if abs(diff[0]) > abs(diff[1]):
            dirs = ['left', 'right']
            diff = diff[0]
        else:
            dirs = ['down', 'up']
            diff = diff[1]
        
        if diff > 0:
            direction = dirs[0]
        else:
            direction = dirs[1]
        
        return direction

class Mutant(GameObject):
    def __init__(self, stage, position):
        #TODO: Implement acceleration, and accelerate at the end of every drag
        #setup animations for each direction
        sheet = pyglet.image.load('mutant%s.png' % stage)
        sequence = pyglet.image.ImageGrid(sheet, 8, 4)
        
        animations = {}
        animations['down'] = pyglet.image.Animation.from_image_sequence([sequence[13], sequence[9], sequence[5], sequence[1]], 0.15)
        animations['up'] = pyglet.image.Animation.from_image_sequence([sequence[12], sequence[8], sequence[4], sequence[0]], 0.15)
        animations['left'] = pyglet.image.Animation.from_image_sequence([sequence[15], sequence[11], sequence[7], sequence[3]], 0.15)
        animations['right'] = pyglet.image.Animation.from_image_sequence([sequence[14], sequence[10], sequence[6], sequence[2]], 0.15)
        
        animations['down_attack'] = pyglet.image.Animation.from_image_sequence([sequence[17], sequence[21], sequence[25],sequence[25],sequence[25],sequence[29]], 0.15)
        animations['up_attack'] = pyglet.image.Animation.from_image_sequence([sequence[16], sequence[20], sequence[24], sequence[24], sequence[24], sequence[28]], 0.15)
        animations['left_attack'] = pyglet.image.Animation.from_image_sequence([sequence[19], sequence[23], sequence[27], sequence[27], sequence[27], sequence[31]], 0.15)
        animations['right_attack'] = pyglet.image.Animation.from_image_sequence([sequence[18], sequence[22], sequence[26], sequence[26], sequence[26], sequence[30]], 0.15)
        
        super(Mutant, self).__init__('mutant%s' % stage, position, direction = 'left', animations = animations)
    
    def vomit(self, dt):
        v = vomit.Vomit(self.direction)
        self.sprite.add(v, 2)
        v.position = (0, 0)
    
    def reset(self, dt):
        self.is_moving = False
        self.is_attacking = False
        
        
    def ai(self, player_rect):
        """
        A fairly simplistic AI
        """
        
        if self.active:
            #close range attack
            if self.sprite.get_rect().intersects(player_rect):
                self.is_moving = False
                self.is_following = True
                
                if not self.is_attacking:
                    self.sprite.image = self.animations['%s_attack' % (self.direction)]
                    clock.schedule_once(self.vomit, 0.3)
                    clock.schedule_once(self.reset, 1.2)
                #self.update_sprite(self.direction_to_point(player_rect.center))
                
                self.is_attacking = True
            
            else:
                self.is_moving = True
                
                if self.is_blocked:
                    #once mutant is blocked (hits wall etc) it will follow player until it hits a wall again. 
                    self.is_following = True
            
                if self.is_following:
                    new_direction = self.direction_to_point(player_rect.center)
                    
                    if new_direction != self.direction:
                        self.update_sprite(new_direction)
                    elif self.is_blocked:
                        self.is_following = False
                        ad = list(self.available_directions)
                        ad.remove(self.direction)
                        self.update_sprite(ad[random.randint(0, 2)])
            
                

class SawSprite(cocos.sprite.Sprite):
    """
    Handles the animation of the saw
    """
    
    def __init__(self, direction = 'down'):
        
        self.direction = direction
        self.animations = {'out': {}, 'in':{},'on': {}}
        self.state = None
        self.next_state = None
        
        #setup saw animations
        #TODO: duplication between out on data
        #TODO: Not sure why everything is in this giant animation list thing
        
        sheet = pyglet.image.load('saw.png')
        sequence = pyglet.image.ImageGrid(sheet, 4, 7)
        
        self.animations['out']['right'] = [pyglet.image.Animation.from_image_sequence(sequence[14:21], 0.1, False), (67, -5), 1]
        self.animations['out']['left'] = [pyglet.image.Animation.from_image_sequence(sequence[21:28], 0.1, False), (-62, -5), 1]
        
        self.animations['on']['right'] = [pyglet.image.Animation.from_image_sequence(sequence[19:21], 0.1), (67, -5), 1]
        self.animations['on']['left'] = [pyglet.image.Animation.from_image_sequence(sequence[26:28], 0.1), (-62, -5), 1]
            
        self.animations['in']['right'] = [pyglet.image.Animation.from_image_sequence([sequence[19], sequence[18], sequence[17], sequence[16], sequence[15], sequence[14]] , 0.1, False), (67, -5), 1]
        self.animations['in']['left'] = [pyglet.image.Animation.from_image_sequence([sequence[26], sequence[25], sequence[24], sequence[23], sequence[22], sequence[21]], 0.1, False), (-62, -5), 1]
        
        self.animations['out']['up'] = [pyglet.image.Animation.from_image_sequence(sequence[0:7], 0.1, False), (0, 30), -1]
        self.animations['out']['down'] = [pyglet.image.Animation.from_image_sequence(sequence[7:14], 0.1, False), (0, -43), 1]
        
        self.animations['on']['up'] = [pyglet.image.Animation.from_image_sequence(sequence[5:7], 0.1), (-1, 30), -1]
        self.animations['on']['down'] = [pyglet.image.Animation.from_image_sequence(sequence[12:14], 0.1), (0, -43), 1]
        
        self.animations['in']['up'] = [pyglet.image.Animation.from_image_sequence([sequence[5], sequence[4], sequence[3], sequence[2], sequence[1], sequence[0]], 0.1, False), (0, 30), -1]
        self.animations['in']['down'] = [pyglet.image.Animation.from_image_sequence([sequence[12], sequence[11], sequence[10], sequence[9], sequence[8], sequence[7]], 0.1, False), (0, -43), 1]
        
        super(SawSprite, self).__init__(self.animations['out'][self.direction][0])
        
        self.animate('out')
        
    
    def animate(self, name):
        if not self.state == 'in':
            self.image = self.animations[name][self.direction][0]
            self.state = name
            
            self.position = self.animations[name][self.direction][1]
            
            if not self.visible:
                self.visible = True
        else:
            if name == 'out':
                self.next_state = name
        
            #TODO: If using a better library, would set frame number and play saw in animation for this direction
            self.visible = False
            
    def get_index(self):
        return self.animations['on'][self.direction][2]
        
    def change_direction(self, direction):
        self.direction = direction
        
        if self.state is not None:
            self.animate('on')
    
    def on_animation_end(self):
        if self.state == 'out':
            self.animate('on')
        if self.state == 'in':
            self.visible = False
            self.state = None
            if self.next_state:
                self.animate(self.next_state)
                self.next_state = None
            
class Player(GameObject):
    """
    The player game object. 'nough said.
    """
    
    def __init__(self, start):
        super(Player, self).__init__('player', position = start)
        
        self.max_speed = 100
        self.energy = 100
        self.hull = 100
        self.base = cocos.euclid.Point2(81, 28)
        self.offset = cocos.euclid.Point2(0, -26)
        self.upgrade_slots = [None, None, None]
        self.saw = None
    
    def saw_on(self):
        if not self.saw:
            self.saw = SawSprite(self.direction)
            self.sprite.add(self.saw, z = self.saw.get_index())
        else:
            self.saw.animate('out')
    
    def saw_off(self):
        if self.saw:
            self.saw.animate('in')
    
    def update_sprite(self, direction):
        self.direction = direction
        self.sprite.image = self.animations[direction]
        self.sprite.position = self.position
        
        if self.saw:
            #AWFUL hack to remove and put sprite back due to the limitations of cocos2d & pyglet (not able to reorder sprite depth? Whaaat?)
            self.sprite.remove(self.saw)
            self.saw.direction = direction
            self.sprite.add(self.saw, z = self.saw.get_index())
            self.saw.change_direction(direction)
        
class Level():
    def __init__(self, name = None):
        
        self.tiles = []
        self.collision = []
        self.mutants = []
        self.start = (0, 0)
        self.sprites = []
        self.data = {}
        self.directions = {'down': (0, -1), 'up': (0, 1), 'left': (-1, 0), 'right': (1, 0) }
        self.tile_offset = (0, 0)
           
        #TODO: load from a sprite sheet instead
        #pyglet.image.atlas
        self.images = {'default': pyglet.resource.image('default.png')}
        
        if name:
            self.load(name)
        else:
            self.lazy_load()
        
    def is_open(self, rect):
        """"
        Uses collision map to work out if rectangle on map is free
        """
        
        for point in [rect.topleft, rect.topright, rect.bottomleft, rect.bottomright]:
            x, y = self.tile_index(point)
            
            if x >= len(self.collision[0]) or x <= 0 or y >= len(self.collision) or y <= 0:
                return False
            elif (self.collision[y][x]):
                return False
        return True
    
    def tile_index(self, point):
        return (int(point[0] / self.data['tile']['width']) + 1, int(point[1] / self.data['tile']['height']) + 1)
    
    def tile_position(self, point):
        return cocos.euclid.Point2(point[0] * self.data['tile']['width'] - self.tile_offset[0], point[1] * self.data['tile']['height'] - self.tile_offset[1])
    
    #TODO: only does if within current tile, expand for surrounding tiles as well so that this is actually usefull
    def direction_to_open_tile(self, direction, point):
        """
        check if the next tile is open
        """
        to_direction = None
        
        #determine on which side the point is
        (i, j) = self.tile_index(point)
        #tile_point = self.tile_position((i,j))
        (di, dj) = self.directions[direction]
        (x, y) = (i + di, j + dj)
        
        if x < len(self.collision[0]) and x > 0 and y < len(self.collision) and y > 0:
            if not(self.collision[y][x]):
                (tx, ty) = self.tile_position((i,j))
                
                #TODO: Couldn't be bothered simplifying to below
                if direction in ('left', 'right'):
                    if ty > point[1]:
                        to_direction = 'up'
                    else:
                        to_direction = 'down'
                elif direction in ('up', 'down'):
                    if tx > point[0]:
                        to_direction = 'right'
                    else:
                        to_direction = 'left'
                
            
        return to_direction
            
        
    def load(self, name):
        """
        Load a level by name. Note: The starting position defines the initial centre of the screen
        """
        print "There is no spoon."
        
    def lazy_load(self, name = 'lazy.map'):
        """
        Loads a level from a text file with a simple grid of numbers
        1 = default, no collision
        """
        
        tiles = {'0': [None, True],
                 '1': ['default', False],
                 '2': ['package1', True],
                 '3': ['package2', True],
                 }
        
        self.tiles = [line.strip('\n').split(',') for line in open(name, 'r')]
        self.collision = list(self.tiles)
        
        self.data = {'tile':{'width': 86, 'height': 36}, 'start':(0, 5)}
        self.tile_offset = (self.data['tile']['width']/2, self.data['tile']['height']/2)
        
        for index, line in enumerate(self.tiles):
            self.tiles[index] = [tiles[value][0] for value in line]
            self.collision[index] = [tiles[value][1] for value in line]
            
        #add a mutant for testing
        self.mutants = [['2', self.tile_position((2,8))], ['2', self.tile_position((16, 3))]]
        
        
    def build(self, batch):
        
        self.start = self.tile_position(self.data['start'])
        
        for y in range(len(self.tiles)):
            for x in range(len(self.tiles[y])):
                if self.tiles[y][x]:
                    sprite = cocos.sprite.Sprite(self.images[self.tiles[y][x]])
                    sprite.position = self.tile_position((x, y))
                    batch.add(sprite, -1)

def main():
    director.init(width = 800, height = 600, vsync = False, resizable = False)
    pyglet.clock.set_fps_limit(300)
    gm = GameScene()
    director.run(gm)

main()
